﻿// #######################################################################################################
// Program Name: CCR Server.exe
// Author: Brice Hilgemann
// Assistance: Professor Blahnik with TcpClient information, MSDN documentation on sockets
// Last Modified: 4/17/2010
// Status: In Progress
// Usage: Creates a Server that has the ability to have multiple clients attached to it in which communcation
//        can occur asynchronously with these and a server can sends information to all of them.
// #######################################################################################################

using System;
using System.Collections.Generic;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Printing;
using System.IO;
using System.IO.Ports;                                   // Standard includes to help with processing and items below
using System.IO.Pipes;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace Server
{
    public partial class Form1 : Form
    {
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//     BEGINNING OF GLOBAL VARIABLES
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

        public delegate void WritetoLogsCallback(string loginfo);                 // Delegate for updating logs in main thread
        public delegate void UpdateLabelsCallback(Label label, string item);      // Delegate for updating labels in main thread


        Socket ServerSocket;                                                      // Socket for the Server
        Socket[] Clients = new Socket[100];                                       // Array of ClientStates for multiple clients
        AsyncCallback CallBack;
        PrintDocument pd;

        byte[] dataholder = new byte[512];                                        // Will act as a buffer for our information
        int connection_count = 0;                                                 // Number of client connections
        int total_connections = 0;                                                // Counts total connections made throughout run-time
        int total_trains = 0;                                                     // Count of trains on the server
        bool expanded = false;                                                    // For form expansion to show logs

        int sensor_count = 29;                                                    // Count of sensors
        int track_count = 31;                                                     // Count of track pieces
        int[] aiu = new int[10];                                                  // Holds AIU information
        int aiu_count = 0;                                                        // Holds AIU count

        PictureBox[] track = new PictureBox[31];                                  // Pictureboxes for the track
        PictureBox[] sensors = new PictureBox[29];                                // Pictureboxes for the sensors

        SerialPort port = new SerialPort();                                       // For COM port communication
        string com_port;                                                          // This will hold the incomming com_port


// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//     BEGINNING OF CODE AND INITIALIZATION
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

        public Form1(string com_selected)
        // ################################################################################################################
        //   Standard Form Initialization
        // ################################################################################################################
        {
            InitializeComponent();

            com_port = com_selected;                             // Set the selected COM port
            label55.Text = com_port;                             // Show the COM port selected in upper right
        }

        public void Form1_Shown(object sender, EventArgs e)
        // ################################################################################################################
        //   Set up defaults at startup when form is first shown
        // ################################################################################################################
        {
            string strHostName = System.Net.Dns.GetHostName();
            IPHostEntry ipEntry = System.Net.Dns.GetHostEntry(strHostName);     // Get current IP Address of system
            IPAddress[] addr = ipEntry.AddressList;

            label5.Text = addr[addr.Length - 1].ToString() + " at Port 10000";
            listBox1.Items.Add(DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Server started with IP " + addr[addr.Length - 1].ToString() + ":10000");

            Form1.ActiveForm.StartPosition = FormStartPosition.CenterScreen;    // Default form settings
            Form1.ActiveForm.MaximizeBox = false;

            initDCC(com_port);                         // Initialization communication with COM port

            SetupListViews();                          // Creates listview columns
            SetupColors();                             // Sets up available colors based off of file
            SetupTrackParts();                         // Sets up track parts based off of file
            SetupTurnouts();                           // Sets up turnouts based off of file
            SetupAIUs();
            SetLabelAttributes();                      // Sets up label attributes
            DefaultImages();                           // Hides and displays proper images

            try
            {
                ServerSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);         // Acts as listener
                IPEndPoint chosen_ip = new IPEndPoint(IPAddress.Any, 10000);                                        // Creates server at default address with port 10000
                ServerSocket.Bind(chosen_ip);                                                                       // Bind the IP Address to listener
                ServerSocket.Listen(5);                                                                             // Create listening queue with queue backlog of 5

                ServerSocket.BeginAccept(new AsyncCallback(HandleClientConnect), null);                             // Set callback to HandleClientConnect if someone tries to connect
            }
            catch (SocketException ex)
            {
                listBox1.Items.Add(DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Socket issue " + ex);         // If something with socket fails
                Form1_FormClosing(null, null);
            }
            catch (ObjectDisposedException)
            {
                MessageBox.Show("HERE");
                Form1_FormClosing(null, null);
            }

            timer1.Interval = 50;                    // Setup a timer of .05 seconds for train updates and AIU processing
            timer1.Enabled = true;

            timer2.Interval = 1000;
            timer2.Enabled = true;
        }


// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//     BEGINNING OF CODE AND SETTING UP DEFAULTS
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

        private void initDCC(string com_port)
        // ################################################################################################################
        //   Sets up communication with the COM port for railroad commands
        // ################################################################################################################
        {
            try
            {
                port = new SerialPort(com_port, 9600, Parity.None, 8, StopBits.One);
                port.BaudRate = 9600;
                port.DataBits = 8;
                port.Parity = Parity.None;                                                  // Set up COM port for communication
                port.StopBits = StopBits.One;                                               //       open it up to write and read
                port.DtrEnable = true;
                port.ReadTimeout = 500;
                port.Handshake = Handshake.None;
                port.DtrEnable = true;
                port.Open();
            }
            catch
            {
                MessageBox.Show("Issue opening up the COM Port - are you sure that you are connected to the railroad?  Try again");
                Form1_FormClosing(null, null);
            }
        }

        public void SetupListViews()
        // ################################################################################################################
        //   Set up defaults for the listviews in the program
        // ################################################################################################################
        {
            SetListViewProperties(listView1);
            listView1.Columns.Add("Train ID", 50, HorizontalAlignment.Center);           //Setup Connection listview
            listView1.Columns.Add("Count", 40, HorizontalAlignment.Center);
            listView1.Columns.Add("IP and Port", 125, HorizontalAlignment.Center);
            CenterListView(listView1);

            SetListViewProperties(listView2);
            listView2.Columns.Add("ID", 30, HorizontalAlignment.Center);                //Setup Turnout listview
            listView2.Columns.Add("State", 50, HorizontalAlignment.Center);
            listView2.Columns.Add("Owner", 50, HorizontalAlignment.Center);
            CenterListView(listView2);

            SetListViewProperties(listView4);
            listView4.Columns.Add("Left", 40, HorizontalAlignment.Center);             //Setup Track listview
            listView4.Columns.Add("Left2", 40, HorizontalAlignment.Center);
            listView4.Columns.Add("Right", 40, HorizontalAlignment.Center);
            listView4.Columns.Add("Right2", 45, HorizontalAlignment.Center);
            listView4.Columns.Add("Turn ID", 50, HorizontalAlignment.Center);
            listView4.Columns.Add("Speed Lmt", 65, HorizontalAlignment.Center);
            listView4.Columns.Add("Trk Lgth", 60, HorizontalAlignment.Center);
            listView4.Columns.Add("Turn St", 50, HorizontalAlignment.Center);
            listView4.Columns.Add("Left Sens", 60, HorizontalAlignment.Center);
            listView4.Columns.Add("Right Sens", 65, HorizontalAlignment.Center);
            listView4.Columns.Add("Nearest TO", 70, HorizontalAlignment.Center);
            CenterListView(listView4);

            SetListViewProperties(listView5);
            listView5.Columns.Add("Train ID", 60, HorizontalAlignment.Center);         //Setup Resources listview
            listView5.Columns.Add("Turn ID", 60, HorizontalAlignment.Center);
            listView5.Columns.Add("Turn St", 60, HorizontalAlignment.Center);
            listView5.Columns.Add("Nearest TO", 70, HorizontalAlignment.Center);
            CenterListView(listView5);

            SetListViewProperties(listView7);
            listView7.Columns.Add("Train ID", 60, HorizontalAlignment.Center);         //Setup Train listview
            listView7.Columns.Add("Owner", 45, HorizontalAlignment.Center);
            listView7.Columns.Add("Track ID", 60, HorizontalAlignment.Center);
            listView7.Columns.Add("Request", 60, HorizontalAlignment.Center);
            listView7.Columns.Add("Blocked", 60, HorizontalAlignment.Center);
            listView7.Columns.Add("Speed", 45, HorizontalAlignment.Center);
            listView7.Columns.Add("Direct.", 60, HorizontalAlignment.Center);
            listView7.Columns.Add("Orient.", 60, HorizontalAlignment.Center);
            listView7.Columns.Add("Light", 45, HorizontalAlignment.Center);
            listView7.Columns.Add("Horn", 40, HorizontalAlignment.Center);
            listView7.Columns.Add("Next Sens", 65, HorizontalAlignment.Center);
            listView7.Columns.Add("Speed Lmt", 65, HorizontalAlignment.Center);
            listView7.Columns.Add("Color", 40, HorizontalAlignment.Center);
            CenterListView(listView7);

            SetListViewProperties(listView9);
            listView9.Columns.Add(" ", 15, HorizontalAlignment.Center);
            listView9.Columns.Add(" ", 15, HorizontalAlignment.Center);                //Setup AIU-Sensor listview
            listView9.Columns.Add(" ", 15, HorizontalAlignment.Center);
            listView9.Columns.Add(" ", 15, HorizontalAlignment.Center);
            listView9.Columns.Add(" ", 15, HorizontalAlignment.Center);
            listView9.Columns.Add(" ", 15, HorizontalAlignment.Center);
            listView9.Columns.Add(" ", 15, HorizontalAlignment.Center);
            listView9.Columns.Add(" ", 15, HorizontalAlignment.Center);
            listView9.Columns.Add(" ", 15, HorizontalAlignment.Center);
            listView9.Columns.Add(" ", 15, HorizontalAlignment.Center);
            listView9.Columns.Add(" ", 15, HorizontalAlignment.Center);
            listView9.Columns.Add(" ", 15, HorizontalAlignment.Center);
            listView9.Columns.Add(" ", 15, HorizontalAlignment.Center);
            listView9.Columns.Add(" ", 15, HorizontalAlignment.Center);
            CenterListView(listView9);
        }

        private void SetListViewProperties(ListView listview)
        // ################################################################################################################
        //   Sets the default settings of a listview for this program
        // ################################################################################################################
        {
            listview.GridLines = true;
            listview.View = View.Details;
            listview.Sorting = SortOrder.None;
        }

        public void CenterListView(ListView listview)
        // ################################################################################################################
        //   Centers the items in a listview
        // ################################################################################################################
        {
            for (int i = 0; i < listview.Columns.Count; i++)
            {
                listview.Columns[i].TextAlign = HorizontalAlignment.Center;
            }
        }

        public void SetupColors()
        // ################################################################################################################
        //   Set up default colors that the user can select from the file
        // ################################################################################################################
        {
            StreamReader sr;

            if (!File.Exists("colors.txt"))                                             // Look for colors.txt file to load colors
            {
                MessageBox.Show("Color file does not exist - starting standard");

                comboBox3.Items.Add("blue");                                            // Set defaults if file can not be found
                comboBox3.Items.Add("red");

                listBox1.Items.Add(DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Default colors set from hard-coded");
            }
            else
            {
                sr = new StreamReader("colors.txt");
                String line;

                while ((line = sr.ReadLine()) != null)                                  // Read color names from file and add to comboBox
                {
                    comboBox3.Items.Add(line);
                }

                sr.Close();

                listBox1.Items.Add(DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Default colors set from file");
            }

            comboBox3.SelectedIndex = 0;                                                // Select first item as default
        }

        public void SetupTrackParts()
        // ################################################################################################################
        //   Set up default track parts to be used on track layout
        // ################################################################################################################
        {
            StreamReader sr;
            ListViewItem lvi;
            ListViewItem.ListViewSubItem lviSub1;
            int counter = 0;

            if (!File.Exists("track.txt"))
            //Open up default track layout for the server to get the track parts and default settings a little more dynamically
            {
                MessageBox.Show("Track file does not exist - Closing program as this is crucial.  Review help to set this up");
                Form1_FormClosing(null, null);
            }
            else
            {
                sr = new StreamReader("track.txt");
                String line;

                while ((line = sr.ReadLine()) != null)
                {
                    string[] dd = line.Split(';');

                    lvi = new ListViewItem(dd[0].ToString());                                  // Add left pointer

                    for (int i = 1; i < dd.Length; i++)
                    {
                        lviSub1 = new ListViewItem.ListViewSubItem(lvi, dd[i].ToString());     // Add left pointer2
                        lvi.SubItems.Add(lviSub1);
                    }

                    listView4.Items.Add(lvi);

                    lvi = new ListViewItem("none");

                    lviSub1 = new ListViewItem.ListViewSubItem(lvi, dd[4].ToString());     // Add turnout id
                    lvi.SubItems.Add(lviSub1);
                    lviSub1 = new ListViewItem.ListViewSubItem(lvi, dd[7].ToString());     // Add turnout state
                    lvi.SubItems.Add(lviSub1);
                    lviSub1 = new ListViewItem.ListViewSubItem(lvi, dd[10].ToString());    // Add Nearest TO
                    lvi.SubItems.Add(lviSub1);

                    listView5.Items.Add(lvi);

                    listBox2.Items.Add(counter.ToString());

                    counter++;
                }

                listBox2.SelectedIndex = 0;              //Select the first index
                sr.Close();                              //Close file reader

                listBox1.Items.Add(DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Default track set from file");
            }
        }

        public void SetupTurnouts()
        // ################################################################################################################
        //   Set up default turnout numbers and scheme for track layout
        // ################################################################################################################
        {
            StreamReader sr;
            ListViewItem lvi;
            ListViewItem.ListViewSubItem lviSub1;

            if (!File.Exists("turnouts.txt"))
            //Open up default turnout layout for the server to get the turnout names and default settings a little more dynamically
            {
                MessageBox.Show("Turnouts file does not exist - starting standard");
                int offset = 4;

                for (int i = 0; i < 6; i++)
                {
                    lvi = new ListViewItem(offset.ToString());

                    lviSub1 = new ListViewItem.ListViewSubItem(lvi, "top");     // Add to listview
                    lvi.SubItems.Add(lviSub1);
                    lviSub1 = new ListViewItem.ListViewSubItem(lvi, "none");    // Add to listview
                    lvi.SubItems.Add(lviSub1);

                    listView2.Items.Add(lvi);
                    offset = (offset * 2);
                }

                listBox1.Items.Add(DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Default turnouts set from program hard-coded");
            }
            else
            {
                sr = new StreamReader("turnouts.txt");
                String line;

                while ((line = sr.ReadLine()) != null)
                {
                    string[] dd = line.Split(';');

                    lvi = new ListViewItem(dd[0].ToString());

                    lviSub1 = new ListViewItem.ListViewSubItem(lvi, dd[1].ToString());     // Add to listview
                    lvi.SubItems.Add(lviSub1);
                    lviSub1 = new ListViewItem.ListViewSubItem(lvi, dd[2].ToString());     // Add to listview
                    lvi.SubItems.Add(lviSub1);

                    listView2.Items.Add(lvi);
                }

                sr.Close();                              //Close file reader

                listBox1.Items.Add(DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Default turnouts set from file");
            }

            byte direction = 1;

            for (int i = 0; i < listView2.Items.Count; i++)
            {
                if (listView2.Items[i].SubItems[1].Text == "bottom")  // See directional byte
                {
                    direction = 0;
                }
                else
                {
                    direction = 1;
                }

                Thread.Sleep(500);                                    // Needed otherwise command doesn't always go through

                SwitchTurnout(Convert.ToUInt32(listView2.Items[i].SubItems[0].Text), direction);     // Initially setup turnouts to structure
            }
        }

        public void SetupAIUs()
        // ################################################################################################################
        //   Set up default aiu address from a file given - if not there, use the default and get first AIU status read
        // ################################################################################################################
        {
            StreamReader sr;

            if (!File.Exists("aiu.txt"))
            //Open up default aiu information to get names to be more dynamic
            {
                MessageBox.Show("AIU file does not exist - starting standard");

                aiu[0] = 50;
                aiu[1] = 54;
                aiu[2] = 58;
                aiu[3] = 62;

                aiu_count = 4;

                listBox1.Items.Add(DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Default aiu infomration set from program hard-coded");
            }
            else
            {
                sr = new StreamReader("aiu.txt");
                String line;
                aiu_count = 0;

                while ((line = sr.ReadLine()) != null)
                {
                    aiu[aiu_count] = Convert.ToInt32(line);
                    aiu_count++;
                }

                listBox1.Items.Add(DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Default aiu information set from file");
            }

            byte[] command = new byte[4];
            int[] value = new int[15];

            //Standard command for AIU messages
            command[0] = 0x9B;

            byte[] buffer = new byte[2];
            byte[] buffer2 = new byte[2];

            for (int i = 0; i < aiu_count; i++)
            {
                command[1] = (byte)aiu[i];

                string junk = port.ReadExisting();

                port.Write(command, 0, 2);

                try
                {
                    port.Read(buffer, 0, 1);             // Reads the hi byte
                    port.Read(buffer2, 0, 1);            // Reads the low byte
                }
                catch { return; }

                for (int j = 0; j < 6; j++)
                {
                    byte test = buffer[0];
                    test = (byte)(test >> j);
                    value[j + 8] = (test & 0x01);
                }

                for (int j = 0; j < 8; j++)
                {
                    byte test = buffer2[0];
                    test = (byte)(test >> j);
                    value[j] = (test & 0x01);
                }

                ListViewItem lvi;
                ListViewItem.ListViewSubItem lviSub1;

                lvi = new ListViewItem(value[0].ToString());

                for (int j = 1; j < 14; j++)
                {
                    lviSub1 = new ListViewItem.ListViewSubItem(lvi, value[j].ToString());     // Add to listview
                    lvi.SubItems.Add(lviSub1);
                }

                listView9.Items.Add(lvi);
            }
        }


// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//     BEGINNING OF CODE AND SETTING UP LABELS/IMAGES
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

        public void SetupBeginningLabels(Label label, int i)
        // ################################################################################################################
        //   Set up beginning labels colors and text
        // ################################################################################################################
        {
            label.BackColor = Color.FromArgb(117, 131, 58);
            label.ForeColor = Color.FromArgb(243, 237, 156);
            label.Text = listView2.Items[i].SubItems[0].Text;
        }

        public void SetupLaterLabels(Label label)
        // ################################################################################################################
        //   Set up later labels colors
        // ################################################################################################################
        {
            label.ForeColor = Color.FromArgb(180, 210, 40);
            label.BackColor = Color.FromArgb(117, 131, 58);
        }

        public void SetLabelAttributes()
        // ################################################################################################################
        //   Set up defaults for the labels and what labels go with what pieces
        // ################################################################################################################
        {
            //Set up label colors and attributes for turnouts - (labels 6-11)
            SetupBeginningLabels(label6, 0);
            SetupBeginningLabels(label7, 1);
            SetupBeginningLabels(label8, 2);
            SetupBeginningLabels(label9, 3);
            SetupBeginningLabels(label11, 4);
            SetupBeginningLabels(label10, 5);

            //Set up the combox attributes
            comboBox1.SelectedIndex = 0;
            comboBox2.SelectedIndex = 0;

            //Set up label colors for track pieces - (labels 23-53)
            SetupLaterLabels(label23);
            SetupLaterLabels(label24);
            SetupLaterLabels(label25);
            SetupLaterLabels(label26);
            SetupLaterLabels(label27);
            SetupLaterLabels(label28);
            SetupLaterLabels(label29);
            SetupLaterLabels(label30);
            SetupLaterLabels(label31);
            SetupLaterLabels(label32);
            SetupLaterLabels(label33);
            SetupLaterLabels(label34);
            SetupLaterLabels(label35);
            SetupLaterLabels(label36);
            SetupLaterLabels(label37);
            SetupLaterLabels(label38);
            SetupLaterLabels(label39);
            SetupLaterLabels(label40);
            SetupLaterLabels(label41);
            SetupLaterLabels(label42);
            SetupLaterLabels(label43);
            SetupLaterLabels(label44);
            SetupLaterLabels(label45);
            SetupLaterLabels(label46);
            SetupLaterLabels(label47);
            SetupLaterLabels(label48);
            SetupLaterLabels(label49);
            SetupLaterLabels(label50);
            SetupLaterLabels(label51);
            SetupLaterLabels(label52);
            SetupLaterLabels(label53);
        }

        public void DefaultImages()
        // ################################################################################################################
        //   Set up default images with certain pictureBoxes
        // ################################################################################################################
        {
            //Top sensors
            sensors[0] = pictureBox15;
            sensors[1] = pictureBox16;
            sensors[2] = pictureBox17;
            sensors[3] = pictureBox18;
            sensors[4] = pictureBox19;
            sensors[5] = pictureBox20;
            sensors[6] = pictureBox21;
            sensors[7] = pictureBox22;
            sensors[8] = pictureBox23;
            sensors[9] = pictureBox24;
            sensors[10] = pictureBox25;
            sensors[11] = pictureBox26;
            sensors[12] = pictureBox27;
            sensors[13] = pictureBox28;
            sensors[14] = pictureBox29;

            //Curve sensors
            sensors[15] = pictureBox30;
            sensors[16] = pictureBox31;
            sensors[17] = pictureBox32;
            sensors[18] = pictureBox33;
            sensors[19] = pictureBox34;
            sensors[20] = pictureBox35;

            //Bottom sensors
            sensors[21] = pictureBox36;
            sensors[22] = pictureBox37;
            sensors[23] = pictureBox38;
            sensors[24] = pictureBox39;
            sensors[25] = pictureBox40;
            sensors[26] = pictureBox41;
            sensors[27] = pictureBox42;
            sensors[28] = pictureBox43;

            for (int i = 0; i < sensor_count; i++)
            {
                sensors[i].Visible = false;
            }

            //Top-level track
            track[0] = pictureBox44;
            track[1] = pictureBox46;
            track[2] = pictureBox47;
            track[3] = pictureBox48;
            track[4] = pictureBox49;
            track[5] = pictureBox50;
            track[6] = pictureBox51;

            //Lower-level track
            track[7] = pictureBox47;
            track[8] = pictureBox52;
            track[9] = pictureBox53;
            track[10] = pictureBox54;
            track[11] = pictureBox55;
            track[12] = pictureBox57;
            track[13] = pictureBox54;
            track[14] = pictureBox58;

            //Bottom-level track
            track[15] = pictureBox59;
            track[16] = pictureBox60;
            track[17] = pictureBox61;
            track[18] = pictureBox62;
            track[19] = pictureBox63;
            track[20] = pictureBox65;
            track[21] = pictureBox66;
            track[22] = pictureBox67;

            //Left curve track
            track[23] = pictureBox45;
            track[24] = pictureBox68;
            track[25] = pictureBox69;
            track[26] = pictureBox62;

            //Right curve track
            track[27] = pictureBox63;
            track[28] = pictureBox71;
            track[29] = pictureBox70;
            track[30] = pictureBox56;

            for (int j = 0; j < track_count; j++)
            {
                track[j].Visible = false;
            }

            //Show default turnout images
            ShowSpecificTurnoutImage(0);
            ShowSpecificTurnoutImage(1);
            ShowSpecificTurnoutImage(2);
            ShowSpecificTurnoutImage(3);
            ShowSpecificTurnoutImage(4);
            ShowSpecificTurnoutImage(5);
        }

        public void ShowTurnoutImages(int turnout_id, PictureBox bottom, PictureBox top)
        // ################################################################################################################
        //   Set up turnout images with certain images pending attributes
        // ################################################################################################################
        {
            if (listView2.Items[turnout_id].SubItems[1].Text == "top")
            {
                top.Image = Properties.Resources.red_turnout;
                bottom.Image = Properties.Resources.off_turnout;
            }
            else
            {
                top.Image = Properties.Resources.off_turnout;
                bottom.Image = Properties.Resources.green_turnout;
            }
        }

        public void ShowSpecificTurnoutImage(int turnout_id)
        // ################################################################################################################
        //   Call function to display specific turnout images according to structures
        // ################################################################################################################
        {
            switch (turnout_id)
            {
                case 0:
                    {
                        ShowTurnoutImages(0, pictureBox3, pictureBox4);
                        break;
                    }
                case 1:
                    {
                        ShowTurnoutImages(1, pictureBox5, pictureBox6);
                        break;
                    }
                case 2:
                    {
                        ShowTurnoutImages(2, pictureBox7, pictureBox8);
                        break;
                    }
                case 3:
                    {
                        ShowTurnoutImages(3, pictureBox9, pictureBox10);
                        break;
                    }
                case 4:
                    {
                        ShowTurnoutImages(4, pictureBox13, pictureBox14);
                        break;
                    }
                case 5:
                    {
                        ShowTurnoutImages(5, pictureBox11, pictureBox12);
                        break;
                    }
            }
        }

        public void UpdateLabels(Label label, string item)
        // ################################################################################################################
        //   Standard delegate to use to update labels on main thread
        // ################################################################################################################
        {
            label.Text = item;
        }


// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//     BEGINNING OF ASYNCHRONOUS CODE AND CALLBACKS
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

        public void HandleClientConnect(IAsyncResult sync_result)
        // ################################################################################################################
        //   Connection to client initiated
        //     - New connection attempt has been made, process and see if a train is available.  Set up sockets, etc.
        // ################################################################################################################
        {
            int available_train_location = 0;

            try
            {
                //See if there is an open train, if there is, give this to the connection
                if (listView1.Items.Count == 0)
                {
                    // There are currently no trains on the track, send message to user
                    Socket TempSocket = (Socket)ServerSocket.EndAccept(sync_result);

                    byte[] response = Encoding.ASCII.GetBytes("0");
                    TempSocket.BeginSend(response, 0, response.Length, SocketFlags.None, new AsyncCallback(ServerSendResponse), TempSocket);

                    ServerSocket.BeginAccept(new AsyncCallback(HandleClientConnect), null);                 // Start another callback to listen for more connections

                    return;
                }
                else
                {
                    // See if any trains are available
                    bool available_trains = false;

                    for (int i = 0; i < listView1.Items.Count; i++)
                    {
                        if (listView1.Items[i].SubItems[2].Text == "Available")
                        {
                            available_train_location = i;
                            available_trains = true;
                        }
                    }

                    if (!available_trains)
                    {
                        // There are currently no available trains, send message to user
                        Socket TempSocket = (Socket)ServerSocket.EndAccept(sync_result);

                        byte[] response = Encoding.ASCII.GetBytes("0");
                        TempSocket.BeginSend(response, 0, response.Length, SocketFlags.None, new AsyncCallback(ServerSendResponse), TempSocket);

                        ServerSocket.BeginAccept(new AsyncCallback(HandleClientConnect), null);                 // Start another callback to listen for more connections

                        return;
                    }
                    else
                    {
                        // Available train has been found, this is what user will use

                        try
                        {
                            Clients[total_connections] = (Socket)ServerSocket.EndAccept(sync_result);               // Process EndAccept and save socket information

                            listView1.Items[available_train_location].SubItems[1].Text = total_connections.ToString();
                            listView1.Items[available_train_location].SubItems[2].Text = Clients[total_connections].RemoteEndPoint.ToString();     // Add to listview

                            for(int j=0; j<listView7.Items.Count; j++)
                            {
                                if(listView7.Items[j].SubItems[0].Text == listView1.Items[available_train_location].SubItems[0].Text)
                                {
                                    listView7.Items[j].SubItems[1].Text = total_connections.ToString();                                      // Mark the owner
                                }
                            }

                            listBox1.Invoke(new WritetoLogsCallback(WritetoLogs), DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Server connected to " + Clients[total_connections].RemoteEndPoint.ToString());

                            string data = "3;" + listView1.Items[available_train_location].SubItems[0].Text + ";" + listView2.Items.Count.ToString();
                            for (int i = 0; i < listView2.Items.Count; i++)
                            {
                                data = data + ";" + listView2.Items[i].SubItems[0].Text;
                            }
                            byte[] data_send = Encoding.ASCII.GetBytes(data);
                            Clients[total_connections].Send(data_send);                                             // Send information to client

                            HandleDatafromClient(Clients[total_connections]);                                       // Call HandleDatafromClient to set up Callback

                            total_connections++;                                                                    // Increase connections
                            connection_count++;

                            label2.Invoke(new UpdateLabelsCallback(UpdateLabels), label2, "Connections: " + connection_count);

                            ServerSocket.BeginAccept(new AsyncCallback(HandleClientConnect), null);                 // Start another callback to listen for more connections
                        }
                        catch
                        {
                            //listBox1.Items.Add(DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Socket issue2 " + ex);         // If something with socket fails
                        }
                    }
                }
            }
            catch
            {
                // Do nothing
            }
        }

        public void HandleDatafromClient(Socket HolderSocket)
        // ################################################################################################################
        //   Information received from Client
        //     - Bytes of data are now in the stream of which are sent from the client, callback was issued and this function
        //       is now called which handles the data
        // ################################################################################################################
        {
            try
            {
                if (CallBack == null)
                {
                    CallBack = new AsyncCallback(HandleDataReceivedfromClient);                         // Set it up to create a callback to HandleDataReceivedfromClient
                }

                HolderSocket.BeginReceive(dataholder, 0, dataholder.Length, SocketFlags.None, CallBack, HolderSocket); // BeginReceive on socket
            }
            catch
            {
                //listBox1.Items.Add(DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Socket issue3 " + ex);           // If something with socket fails
            }
        }

        public void HandleDataReceivedfromClient(IAsyncResult sync_result)
        // ################################################################################################################
        //   Information received from Client - now handle the data
        //     - Bytes of data are now in the stream of which are sent from the client, callback was issued and this function
        //       is now called which handles the data
        // ################################################################################################################
        {
            Socket TempSocket = null;
            string[] data = new string[20];

            try
            {
                TempSocket = (Socket)sync_result.AsyncState;                    // Get information sent in temporary socket

                int received = TempSocket.EndReceive(sync_result);              // Store it in an int to then decode

                if (received > 0)
                {
                    string received_info = Encoding.ASCII.GetString(dataholder, 0, received);    // Convert received to string
                    data = received_info.Split(';');                                             //Check received_info for a message

                    listBox1.Invoke(new WritetoLogsCallback(WritetoLogs), DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - From Client " + TempSocket.RemoteEndPoint.ToString() + " " + received_info);    //Output string received

                    // ****************************************************************************************************
                    //     PROCESS REQUEST FROM CLIENT
                    // ****************************************************************************************************

                    ProcessClientRequest(data);

                    HandleDatafromClient(TempSocket);
                }
                else
                {
                    CloseSocket(TempSocket);
                }
            }
            catch
            {
                CloseSocket(TempSocket);
            }
        }

        private void ProcessClientRequest(string[] data)
        // ################################################################################################################
        //   Function used if the server is sending a message to the client - not included right now
        // ################################################################################################################
        {
            Monitor.Enter(listView5);
            Monitor.Enter(listView2);
            Monitor.Enter(listView7);

            switch (data[0])
            {
                case "1":
                // **************************************************************************************************************
                //     User wants to increase speed or go forward - check and update if appropriate
                // **************************************************************************************************************
                    {
                        for (int i = 0; i < listView7.Items.Count; i++)
                        {
                            if (data[1] == listView7.Items[i].SubItems[0].Text)     //Train ID
                            {
                                if ((Convert.ToInt32(listView7.Items[i].SubItems[5].Text) < 80) && (Convert.ToInt32(listView7.Items[i].SubItems[5].Text) < Convert.ToInt32(listView7.Items[i].SubItems[11].Text)))     // See if < 80, add 10 to it
                                {
                                    //Increase the train by 10 if it is less than 80 and Track Speed is greater than speed
                                    if (listView7.Items[i].SubItems[4].Text == "false")
                                    {
                                        //Train is not blocked, change it
                                        listView7.Items[i].SubItems[5].Text = (Convert.ToInt32(listView7.Items[i].SubItems[5].Text) + 10).ToString();
                                    }
                                }

                                if (Convert.ToInt32(listView7.Items[i].SubItems[5].Text) == 0 && listView7.Items[i].SubItems[4].Text == "true" && listView7.Items[i].SubItems[6].Text == "Reverse")
                                {
                                    listView7.Items[i].SubItems[6].Text = "Forward";
                                    listView7.Items[i].SubItems[5].Text = "10";

                                    TrainForward(Convert.ToInt32(data[1]), Convert.ToInt32(listView7.Items[i].SubItems[5].Text));

                                    Monitor.Exit(listView7);
                                    Monitor.Exit(listView2);
                                    Monitor.Exit(listView5);

                                    //Put in new request to get track in opposite direction
                                    RequestNextTrack(data[1]);

                                    return;
                                }

                                if (Convert.ToInt32(listView7.Items[i].SubItems[5].Text) < 0)
                                {
                                    TrainBackward(Convert.ToInt32(data[1]), Convert.ToInt32(listView7.Items[i].SubItems[5].Text));
                                    
                                    // Train is moving backward at speed
                                    if (listView7.Items[i].SubItems[6].Text != "Reverse")
                                    {
                                        listView7.Items[i].SubItems[6].Text = "Reverse";

                                        Monitor.Exit(listView7);
                                        Monitor.Exit(listView2);
                                        Monitor.Exit(listView5);

                                        //Put in new request to get track in opposite direction
                                        RequestNextTrack(data[1]);

                                        return;
                                    }
                                }
                                else
                                {
                                    TrainForward(Convert.ToInt32(data[1]), Convert.ToInt32(listView7.Items[i].SubItems[5].Text));

                                    // Train is moving forward at speed
                                    if (listView7.Items[i].SubItems[6].Text != "Forward")
                                    {
                                        listView7.Items[i].SubItems[6].Text = "Forward";

                                        Monitor.Exit(listView7);
                                        Monitor.Exit(listView2);
                                        Monitor.Exit(listView5);

                                        //Put in new request to get track in opposite direction
                                        RequestNextTrack(data[1]);

                                        return;
                                    }
                                }
                            }
                        }

                        Monitor.Exit(listView7);
                        Monitor.Exit(listView2);
                        Monitor.Exit(listView5);

                        return;
                    }
                case "2":
                // **************************************************************************************************************
                //    User wants to decrease speed or go backward - check and update if appropriate
                // **************************************************************************************************************
                    {
                        for (int i = 0; i < listView7.Items.Count; i++)
                        {
                            if (data[1] == listView7.Items[i].SubItems[0].Text)     //Train ID
                            {
                                if ((Math.Abs(Convert.ToInt32(listView7.Items[i].SubItems[5].Text)) < 80) && (Convert.ToInt32(listView7.Items[i].SubItems[5].Text) > (Convert.ToInt32(listView7.Items[i].SubItems[11].Text) * -1)))      //Decrease the train by 10 if it is less than 80
                                {
                                    if (listView7.Items[i].SubItems[4].Text == "false")
                                    {
                                        //Train is not blocked, change speed
                                        listView7.Items[i].SubItems[5].Text = (Convert.ToInt32(listView7.Items[i].SubItems[5].Text) - 10).ToString();
                                    }
                                }

                                if (Convert.ToInt32(listView7.Items[i].SubItems[5].Text) == 0 && listView7.Items[i].SubItems[4].Text == "true" && listView7.Items[i].SubItems[6].Text == "Forward")
                                {
                                    //User is blocked and has now just changed directions = increase speed and change directions

                                    listView7.Items[i].SubItems[6].Text = "Reverse";
                                    listView7.Items[i].SubItems[5].Text = "-10";

                                    TrainBackward(Convert.ToInt32(data[1]), Convert.ToInt32(listView7.Items[i].SubItems[5].Text));

                                    Monitor.Exit(listView7);
                                    Monitor.Exit(listView2);
                                    Monitor.Exit(listView5);

                                    //Put in new request to get track in opposite direction
                                    RequestNextTrack(data[1]);

                                    return;
                                }

                                if (Convert.ToInt32(listView7.Items[i].SubItems[5].Text) < 0)
                                {
                                    TrainBackward(Convert.ToInt32(data[1]), Convert.ToInt32(listView7.Items[i].SubItems[5].Text));

                                    // Train is moving backward at speed
                                    if (listView7.Items[i].SubItems[6].Text != "Reverse")
                                    {
                                        listView7.Items[i].SubItems[6].Text = "Reverse";

                                        Monitor.Exit(listView7);
                                        Monitor.Exit(listView2);
                                        Monitor.Exit(listView5);

                                        //Put in new request to get track in opposite direction
                                        RequestNextTrack(data[1]);

                                        return;
                                    }
                                }
                                else
                                {
                                    TrainForward(Convert.ToInt32(data[1]), Convert.ToInt32(listView7.Items[i].SubItems[5].Text));

                                    // Train is moving forward at speed
                                    if (listView7.Items[i].SubItems[6].Text != "Forward")
                                    {
                                        listView7.Items[i].SubItems[6].Text = "Forward";

                                        Monitor.Exit(listView7);
                                        Monitor.Exit(listView2);
                                        Monitor.Exit(listView5);

                                        //Put in new request to get track in opposite direction
                                        RequestNextTrack(data[1]);

                                        return;
                                    }
                                }
                            }
                        }

                        Monitor.Exit(listView7);
                        Monitor.Exit(listView2);
                        Monitor.Exit(listView5);

                        return;
                    }
                case "3":
                // **************************************************************************************************************
                //    User wants to stop the train
                // **************************************************************************************************************
                    {
                        for (int i = 0; i < listView7.Items.Count; i++)
                        {
                            if (data[1] == listView7.Items[i].SubItems[0].Text)     //Train ID
                            {
                                listView7.Items[i].SubItems[5].Text = "0";

                                TrainForward(Convert.ToInt32(data[1]), Convert.ToInt32(listView7.Items[i].SubItems[5].Text));
                            }
                        }

                        Monitor.Exit(listView7);
                        Monitor.Exit(listView2);
                        Monitor.Exit(listView5);

                        return;
                    }
                case "4":
                // **************************************************************************************************************
                //    User wants to Change direction to forward
                // **************************************************************************************************************
                    {
                        for (int i = 0; i < listView7.Items.Count; i++)
                        {
                            if (data[1] == listView7.Items[i].SubItems[0].Text)     //Train ID
                            {
                                if (listView7.Items[i].SubItems[6].Text == "Reverse")
                                {
                                    listView7.Items[i].SubItems[6].Text = "Forward";
                                    listView7.Items[i].SubItems[5].Text = Math.Abs(Convert.ToInt32(listView7.Items[i].SubItems[5].Text)).ToString();

                                    TrainForward(Convert.ToInt32(data[1]), Convert.ToInt32(listView7.Items[i].SubItems[5].Text));

                                    Monitor.Exit(listView7);
                                    Monitor.Exit(listView2);
                                    Monitor.Exit(listView5);

                                    RequestNextTrack(data[1]);

                                    return;
                                }
                            }
                        }

                        Monitor.Exit(listView7);
                        Monitor.Exit(listView2);
                        Monitor.Exit(listView5);

                        return;
                    }
                case "5":
                // **************************************************************************************************************
                //    User wants to Change direction to backward
                // **************************************************************************************************************
                    {
                        for (int i = 0; i < listView7.Items.Count; i++)
                        {
                            if (data[1] == listView7.Items[i].SubItems[0].Text)     //Train ID
                            {
                                if (listView7.Items[i].SubItems[6].Text == "Forward")
                                {
                                    listView7.Items[i].SubItems[6].Text = "Reverse";
                                    listView7.Items[i].SubItems[5].Text = (Convert.ToInt32(listView7.Items[i].SubItems[5].Text) * -1).ToString();

                                    TrainBackward(Convert.ToInt32(data[1]), Convert.ToInt32(listView7.Items[i].SubItems[5].Text));

                                    Monitor.Exit(listView7);
                                    Monitor.Exit(listView2);
                                    Monitor.Exit(listView5);

                                    RequestNextTrack(data[1]);

                                    return;
                                }
                            }
                        }

                        Monitor.Exit(listView7);
                        Monitor.Exit(listView2);
                        Monitor.Exit(listView5);

                        return;
                    }
                case "6":
                // **************************************************************************************************************
                //    User wants to Toggle a turnout
                // **************************************************************************************************************
                    {
                        byte direction = 1;
                        int turnout_id = Convert.ToInt32(data[1]);

                        if (listView2.Items[turnout_id].SubItems[2].Text == "none")       // See if user can switch turnout
                        {
                            if (listView2.Items[turnout_id].SubItems[1].Text == "top")
                            {
                                listView2.Items[turnout_id].SubItems[1].Text = "bottom";
                                direction = 0;
                            }
                            else
                            {
                                listView2.Items[turnout_id].SubItems[1].Text = "top";
                            }

                            SwitchTurnout(Convert.ToUInt32(listView2.Items[turnout_id].SubItems[0].Text), direction);   // Call to switch turnout
                            ShowSpecificTurnoutImage(turnout_id);                                                       // Show new image
                        }

                        Monitor.Exit(listView7);
                        Monitor.Exit(listView2);
                        Monitor.Exit(listView5);

                        return;
                    }
                case "7":
                // **************************************************************************************************************
                //    User wants to Add a train
                // **************************************************************************************************************
                    {
                        //Check if there is a turnout associated with resource
                        if (listView5.Items[Convert.ToInt32(data[5])].SubItems[1].Text == "none")
                        {
                            //No turnout for track, check if resource is open
                            if (listView5.Items[Convert.ToInt32(data[5])].SubItems[0].Text == "none")
                            {
                                //No trains associated with track
                            }
                            else
                            {
                                Monitor.Exit(listView7);
                                Monitor.Exit(listView2);
                                Monitor.Exit(listView5);
                                MessageBox.Show("Track is in use - Please select another");
                                return;
                            }
                        }
                        else
                        {
                            //Verify that track is unused and turnout is turned in right direction
                            if (listView5.Items[Convert.ToInt32(data[5])].SubItems[0].Text == "none" &&
                                listView2.Items[Convert.ToInt32(listView5.Items[Convert.ToInt32(data[5])].SubItems[1].Text)].SubItems[1].Text == listView5.Items[Convert.ToInt32(data[5])].SubItems[2].Text)
                            {
                                //No trains associated with track and turnout
                                listView2.Items[Convert.ToInt32(listView5.Items[Convert.ToInt32(data[5])].SubItems[1].Text)].SubItems[2].Text = data[1];     // Set TrainID for this Turnout ownership
                            }
                            else
                            {
                                Monitor.Exit(listView7);
                                Monitor.Exit(listView2);
                                Monitor.Exit(listView5);
                                MessageBox.Show("Track is in use - Please select another");
                                return;
                            }
                        }

                        //Add Train to Trains and connections
                        ListViewItem lvi;
                        ListViewItem.ListViewSubItem lviSub1;                                                   // Create listview items

                        lvi = new ListViewItem(data[1]);
                        lviSub1 = new ListViewItem.ListViewSubItem(lvi, "");                                    // Add to listview
                        lvi.SubItems.Add(lviSub1);
                        lviSub1 = new ListViewItem.ListViewSubItem(lvi, "Available");                           // Add to listview
                        lvi.SubItems.Add(lviSub1);

                        listView1.Items.Add(lvi);

                        listView1.Items[listView1.Items.Count - 1].BackColor = Color.FromName(data[4]);         // Change backcolor
                        if (data[4] == "yellow")
                        {
                            listView1.Items[listView1.Items.Count - 1].ForeColor = Color.Black;                 // Change forecolor
                        }
                        else
                        {
                            listView1.Items[listView1.Items.Count - 1].ForeColor = Color.White;
                        }

                        //Add Train Information
                        lvi = new ListViewItem(data[1]);                             // Train ID
                        lviSub1 = new ListViewItem.ListViewSubItem(lvi, "none");     // Ownership
                        lvi.SubItems.Add(lviSub1);
                        lviSub1 = new ListViewItem.ListViewSubItem(lvi, "none");     // Track ID
                        lvi.SubItems.Add(lviSub1);
                        lviSub1 = new ListViewItem.ListViewSubItem(lvi, data[5]);    // Request
                        lvi.SubItems.Add(lviSub1);
                        lviSub1 = new ListViewItem.ListViewSubItem(lvi, "false");    // Blocked
                        lvi.SubItems.Add(lviSub1);
                        lviSub1 = new ListViewItem.ListViewSubItem(lvi, "0");        // Speed
                        lvi.SubItems.Add(lviSub1);
                        lviSub1 = new ListViewItem.ListViewSubItem(lvi, data[2]);    // Direction
                        lvi.SubItems.Add(lviSub1);
                        lviSub1 = new ListViewItem.ListViewSubItem(lvi, data[3]);    // Orientation
                        lvi.SubItems.Add(lviSub1);
                        lviSub1 = new ListViewItem.ListViewSubItem(lvi, "false");    // Light
                        lvi.SubItems.Add(lviSub1);
                        lviSub1 = new ListViewItem.ListViewSubItem(lvi, "false");    // Horn
                        lvi.SubItems.Add(lviSub1);

                        //Look at direction and orientation to determine next sensor
                        if ((data[3] == "Left" && data[2] == "Forward") || (data[3] == "Right" && data[2] == "Reverse"))
                        {
                            //Train should be going left, pick the right sensor for now
                            lviSub1 = new ListViewItem.ListViewSubItem(lvi, listView4.Items[Convert.ToInt32(data[5])].SubItems[9].Text);      // Next Sensor
                            lvi.SubItems.Add(lviSub1);
                        }
                        else
                        {
                            //Train should be going right, pick the left sensor for now
                            lviSub1 = new ListViewItem.ListViewSubItem(lvi, listView4.Items[Convert.ToInt32(data[5])].SubItems[8].Text);      // Next Sensor
                            lvi.SubItems.Add(lviSub1);
                        }

                        lviSub1 = new ListViewItem.ListViewSubItem(lvi, listView4.Items[Convert.ToInt32(data[5])].SubItems[5].Text);     // Speed Limit
                        lvi.SubItems.Add(lviSub1);
                        lviSub1 = new ListViewItem.ListViewSubItem(lvi, data[4]);    // Color
                        lvi.SubItems.Add(lviSub1);

                        listView7.Items.Add(lvi);

                        //Remove color taken
                        comboBox3.Items.Remove(comboBox3.SelectedItem);
                        try
                        {
                            comboBox3.SelectedIndex = 0;
                        }
                        catch { }

                        //Add train to delete train list
                        listBox3.Items.Add("Train " + data[1]);

                        listView5.Items[Convert.ToInt32(data[5])].SubItems[0].Text = data[1];    // Set TrainID for Resource ownership
                        total_trains++;

                        label1.Invoke(new UpdateLabelsCallback(UpdateLabels), label1, "Trains: " + total_trains);

                        //Exit monitors
                        Monitor.Exit(listView7);
                        Monitor.Exit(listView2);
                        Monitor.Exit(listView5);

                        listBox1.Invoke(new WritetoLogsCallback(WritetoLogs), DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Train " + data[0] + " added to railroad");                //Put in a sensor hit on the sensor away from where it is going
                        SensorHitChange(data[1]);

                        return;
                    }
                case "8":
                // **************************************************************************************************************
                //    User wants to remove a train
                // **************************************************************************************************************
                    {
                        //Nothing for this - implemented in another area

                        break;
                    }
                case "9":
                // **************************************************************************************************************
                //    User wants to toggle the light
                // **************************************************************************************************************
                    {
                        for (int i = 0; i < listView7.Items.Count; i++)
                        {
                            if (listView7.Items[i].SubItems[8].Text == "false")
                            {
                                listView7.Items[i].SubItems[8].Text = "true";
                            }
                            else
                            {
                                listView7.Items[i].SubItems[8].Text = "false";
                            }
                        }

                        Monitor.Exit(listView7);
                        Monitor.Exit(listView2);
                        Monitor.Exit(listView5);

                        return;
                    }
                case "10":
                // **************************************************************************************************************
                //    User wants to toggle the horn
                // **************************************************************************************************************
                    {
                        for (int i = 0; i < listView7.Items.Count; i++)
                        {
                            if (listView7.Items[i].SubItems[9].Text == "false")
                            {
                                listView7.Items[i].SubItems[9].Text = "true";
                            }
                            else
                            {
                                listView7.Items[i].SubItems[9].Text = "false";
                            }
                        }

                        Monitor.Exit(listView7);
                        Monitor.Exit(listView2);
                        Monitor.Exit(listView5);

                        return;
                    }
            }
        }

        public void ServerSendResponse(IAsyncResult sync_result)
        // ################################################################################################################
        //   Function used if the server is sending a message to the client - not included right now
        // ################################################################################################################
        {
            Socket TempSocket = null;

            HandleDatafromClient(TempSocket);
        }

        private void timer1_Tick(object sender, EventArgs e)
        // ################################################################################################################
        //   Utilize a timer to check AIU's to see where trains are moving on the track
        // ################################################################################################################
        {
            CheckAIU();                                                        // Check AIU's and change images, perform requests if hit, etc.
        }

        private void timer2_Tick(object sender, EventArgs e)
        // ################################################################################################################
        //   Utilize a timer to check blocked trains and send updates to clients
        // ################################################################################################################
        {
            for (int i = 0; i < listView7.Items.Count; i++)                         // Go through connections and send updates
            {
                if (listView7.Items[i].SubItems[4].Text == "true")
                {
                    RequestNextTrack(listView7.Items[i].SubItems[0].Text);     // Train is blocked - see if it can now be unblocked
                }

                if (listView7.Items[i].SubItems[1].Text != "none")
                {
                    try
                    {
                        string data = "1;" +                                                                          // Update
                            listView7.Items[i].SubItems[0].Text + ";" +                                               // Train ID
                            listView7.Items[i].SubItems[5].Text + ";" +                                               // Speed
                            listView7.Items[i].SubItems[6].Text + ";" +                                               // Direction
                            listView7.Items[i].SubItems[8].Text + ";" +                                               // Light
                            listView7.Items[i].SubItems[9].Text + ";" +                                               // Horn
                            listView7.Items[i].SubItems[4].Text + ";" +                                               // Blocked Status
                            listView5.Items[Convert.ToInt32(listView7.Items[i].SubItems[2].Text)].SubItems[3].Text + ";" +   // Nearest TO
                            listView2.Items.Count;                                                                    // Nbr of TO's

                        for (int j = 0; j < listView2.Items.Count; j++)
                        {
                            data = data + ";" + listView2.Items[j].SubItems[1].Text;                                  // Each TO Status
                        }

                        byte[] data_send = Encoding.ASCII.GetBytes(data);
                        Clients[Convert.ToInt32(listView1.Items[i].SubItems[1].Text)].Send(data_send);
                    }
                    catch
                    {
                        //Did client disconnect?
                    }
                }
            }
        }


// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//     BEGINNING OF TURNOUT SPECIFIC CODE AND EVENTS
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

        private void pictureBox4_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   User has clicked the upper Turnout 0
        // ################################################################################################################
        {
            CheckTurnout(0, "top", pictureBox3, pictureBox4);
        }
        private void pictureBox3_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   User has clicked the lower Turnout 0
        // ################################################################################################################
        {
            CheckTurnout(0, "bottom", pictureBox3, pictureBox4);
        }

        private void pictureBox6_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   User has clicked the upper Turnout 1
        // ################################################################################################################
        {
            CheckTurnout(1, "top", pictureBox5, pictureBox6);
        }
        private void pictureBox5_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   User has clicked the lower Turnout 1
        // ################################################################################################################
        {
            CheckTurnout(1, "bottom", pictureBox5, pictureBox6);
        }

        private void pictureBox8_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   User has clicked the upper Turnout 2
        // ################################################################################################################
        {
            CheckTurnout(2, "top", pictureBox7, pictureBox8);
        }
        private void pictureBox7_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   User has clicked the lower Turnout 2
        // ################################################################################################################
        {
            CheckTurnout(2, "bottom", pictureBox7, pictureBox8);
        }

        private void pictureBox10_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   User has clicked the upper Turnout 3
        // ################################################################################################################
        {
            CheckTurnout(3, "top", pictureBox9, pictureBox10);
        }
        private void pictureBox9_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   User has clicked the lower Turnout 3
        // ################################################################################################################
        {
            CheckTurnout(3, "bottom", pictureBox9, pictureBox10);
        }

        private void pictureBox14_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   User has clicked the upper Turnout 4
        // ################################################################################################################
        {
            CheckTurnout(4, "top", pictureBox13, pictureBox14);
        }
        private void pictureBox13_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   User has clicked the lower Turnout 4
        // ################################################################################################################
        {
            CheckTurnout(4, "bottom", pictureBox13, pictureBox14);
        }

        private void pictureBox12_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   User has clicked the upper Turnout 5
        // ################################################################################################################
        {
            CheckTurnout(5, "top", pictureBox11, pictureBox12);
        }
        private void pictureBox11_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   User has clicked the lower Turnout 5
        // ################################################################################################################
        {
            CheckTurnout(5, "bottom", pictureBox11, pictureBox12);
        }

        public void CheckTurnout(int turnout_id, string location, PictureBox bottom, PictureBox top)
        // ################################################################################################################
        //   Main function for checking and calling to switch turnouts if applicable
        // ################################################################################################################
        {
            //Enter monitor
            Monitor.Enter(listView2);

            byte direction = 1;

            if (listView2.Items[turnout_id].SubItems[2].Text == "none" && listView2.Items[turnout_id].SubItems[1].Text != location)   // See if user wants to switch
            {
                listView2.Items[turnout_id].SubItems[1].Text = location;                      // Get it's location

                if (location == "bottom")                                                     // Get directional byte
                {
                    direction = 0;
                }

                SwitchTurnout(Convert.ToUInt32(listView2.Items[turnout_id].SubItems[0].Text), direction);   // Call to switch turnout

                ShowTurnoutImages(turnout_id, bottom, top);                                                 // Show the corresponding turnout image

                listBox1.Invoke(new WritetoLogsCallback(WritetoLogs), DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Turnout " + listView2.Items[turnout_id].SubItems[0].Text + " changed to " + listView2.Items[turnout_id].SubItems[1].Text);
            }

            Monitor.Exit(listView2);
            //Exit monitor
        }

        public void SwitchTurnout(uint turnout, byte direction)
        // ################################################################################################################
        //   Communicates with the serial port to switch the turnout
        // ################################################################################################################
        {
            byte[] command = new byte[6];
            byte[] buffer = new byte[2];

            command[0] = 0xAD;                                          // Standard accessory packet command
            command[1] = 0;                                             // High address, here set to 0 to be simple
            command[2] = (byte)turnout;                                 // Address of turnout
            command[3] = (byte)(04 - direction);                        // Direction
            command[4] = 0;                                             // Default data

            try
            {
                port.Write(command, 0, 5);

                port.Read(buffer, 0, 1);
            }
            catch { }
        }


// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//     BEGINNING OF ADDING AND DELETING TRAINS SPECIFIC CODE AND EVENTS
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

        private void pictureBox72_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   User clicked that they want to add a train, attempt to add it according to selected
        // ################################################################################################################
        {
            string[] data = new string[10];
            bool bad_train = false;
            int train_id;

            //Verify that train is in an int
            try
            {
                train_id = Convert.ToInt32(textBox2.Text);

                //Verify that train id is within 1-500
                if (train_id > 0 && train_id < 501)
                {
                    //Verify that train id is not in use
                    for (int i = 0; i < listView1.Items.Count; i++)
                    {
                        if (listView1.Items[i].SubItems[0].Text == train_id.ToString())
                        {
                            bad_train = true;
                        }
                    }
                }
                else
                {
                    bad_train = true;
                }
            }
            catch
            {
                MessageBox.Show("Train is already in use or Short ID not within 1-500");
                return;
            }

            if (bad_train == true)
            {
                MessageBox.Show("Train is already in use or Short ID not within 1-500");
                return;
            }

            //Verify that colors are valid and one is chosen
            try
            {
                if (comboBox3.Text.Length == 0)
                {
                    MessageBox.Show("No more trains can be added at this time");
                    return;
                }
            }
            catch
            {
                MessageBox.Show("No more trains can be added at this time");
                return;
            }

            data[0] = "7";                                // Command to add a train
            data[1] = train_id.ToString();                // Train ID
            data[2] = comboBox1.Text;                     // Direction
            data[3] = comboBox2.Text;                     // Facing
            data[4] = comboBox3.Text;                     // Color
            data[5] = listBox2.SelectedIndex.ToString();

            ProcessClientRequest(data);
        }

        private void pictureBox76_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   Remove trains 'button' clicked - remove selected train
        // ################################################################################################################
        {
            if (listBox3.SelectedItems.Count == 0)
            {
                MessageBox.Show("No train selected to remove or there are no trains");
            }
            else
            {
                string[] data = new string[5];
                data = listBox3.Items[listBox3.SelectedIndex].ToString().Split(' ');

                for (int i = 0; i < listView1.Items.Count; i++)
                {
                    try
                    {
                        if (listView1.Items[i].SubItems[0].Text == data[1])
                        {
                            try
                            {
                                if (listView1.Items[i].SubItems[2].Text != "Available")
                                {
                                    //Owner of train, disconnect them and delete train
                                    Socket TempSocket = Clients[Convert.ToInt32(listView1.Items[i].SubItems[1].Text)];

                                    byte[] response = Encoding.ASCII.GetBytes("2");
                                    TempSocket.BeginSend(response, 0, response.Length, SocketFlags.None, new AsyncCallback(ServerSendResponse), Clients[Convert.ToInt32(listView1.Items[i].SubItems[1].Text)]);
                                }
                            }
                            catch { }

                            // Remove ownership of train and then delete from listBox and listViews
                            DeallocateTrack(i);
                            DeallocateRequestTrack(i);

                            listView1.Items[i].Remove();

                            total_trains--;

                            TrainForward(Convert.ToInt32(listView7.Items[i].SubItems[0].Text), 0);

                            listBox3.Items.RemoveAt(listBox3.SelectedIndex);
                            comboBox3.Items.Add(listView7.Items[i].SubItems[12].Text);
                            listView7.Items[i].Remove();

                            label1.Invoke(new UpdateLabelsCallback(UpdateLabels), label1, "Trains: " + total_trains);
                        }
                    }
                    catch
                    {
                        //Ignore as listBox probably has no more trains
                        MessageBox.Show("Unable to remove train");
                    }
                }
            }
        }


// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//     BEGINNING OF CODE FOR TRAIN CHANGES TO PREVENT COLLISIONS AND OWNERSHIP ISSUES
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

        public int GetTrainLocation(string train_id)
        // ################################################################################################################
        //   Gets the trains current location
        // ################################################################################################################
        {
            int train_location = 0;

            //Find the train item associated with this train_id
            for (int i = 0; i < listView7.Items.Count; i++)
            {
                if (listView7.Items[i].SubItems[0].Text == train_id)
                {
                    train_location = i;
                }
            }

            return train_location;
        }

        private void DeallocateRequestTrack(int train_location)
        // ################################################################################################################
        //   Deallocate the current 'Requested' track at this current train location
        // ################################################################################################################
        {
            if (listView7.Items[train_location].SubItems[3].Text != "none")
            {
                // Need to deallocate the old existing track that is in 'Request'
                track[Convert.ToInt32(listView7.Items[train_location].SubItems[3].Text)].Visible = false;                            // Hide track image

                if (listView5.Items[Convert.ToInt32(listView7.Items[train_location].SubItems[3].Text)].SubItems[1].Text != "none")   // Check TO Ownership
                {
                    //Open up the turnout ownership also
                    listView2.Items[Convert.ToInt32(listView5.Items[Convert.ToInt32(listView7.Items[train_location].SubItems[3].Text)].SubItems[1].Text)].SubItems[2].Text = "none";
                }

                listView5.Items[Convert.ToInt32(listView7.Items[train_location].SubItems[3].Text)].SubItems[0].Text = "none";        // Officially remove from Resource
                listView7.Items[train_location].SubItems[3].Text = "none";                                                           // Officially remove from Train
            }
        }

        private void DeallocateTrack(int train_location)
        // ################################################################################################################
        //   Deallocate the current 'Track ID' track at this current train location
        // ################################################################################################################
        {
            if (listView7.Items[train_location].SubItems[2].Text != "none")                                                          // See if user owns a 'Track ID'
            {
                // Need to deallocate the old existing track that is in 'Track ID'
                track[Convert.ToInt32(listView7.Items[train_location].SubItems[2].Text)].Visible = false;                            // Hide track image

                if (listView5.Items[Convert.ToInt32(listView7.Items[train_location].SubItems[2].Text)].SubItems[1].Text != "none")   // Check TO Ownership
                {
                    //Open up the turnout ownership also
                    listView2.Items[Convert.ToInt32(listView5.Items[Convert.ToInt32(listView7.Items[train_location].SubItems[2].Text)].SubItems[1].Text)].SubItems[2].Text = "none";
                }

                listView5.Items[Convert.ToInt32(listView7.Items[train_location].SubItems[2].Text)].SubItems[0].Text = "none";        // Officially remove from Resource
                listView7.Items[train_location].SubItems[2].Text = "none";                                                           // Officially remove from Train
            }
        }

        private void GetRequestedTrack(int train_location, int current_track, ref string track_piece, ref string next_sensor)
        // ################################################################################################################
        //   Get the requested track and next sensor train should be going
        // ################################################################################################################
        {
            //Get the next requested track and see if it is available and turnout is in the right direction if applicable
            if ((listView7.Items[train_location].SubItems[7].Text == "Left" && listView7.Items[train_location].SubItems[6].Text == "Forward") ||
                (listView7.Items[train_location].SubItems[7].Text == "Right" && listView7.Items[train_location].SubItems[6].Text == "Reverse"))
            {
                //Train should be going left, pick the next track and sensor
                next_sensor = listView4.Items[Convert.ToInt32(listView7.Items[train_location].SubItems[2].Text)].SubItems[8].Text;

                if (listView4.Items[current_track].SubItems[1].Text == "none")       // Get what track they are currently on
                {
                    //Only one track piece can be selected for this direction
                    track_piece = listView4.Items[current_track].SubItems[0].Text;
                }
                else
                {
                    //Multiple pieces could be available - look at turnout state to then see which track it should be going to
                    int left2 = Convert.ToInt32(listView4.Items[current_track].SubItems[1].Text);
                    int turnout = Convert.ToInt32(listView5.Items[left2].SubItems[1].Text);

                    if(listView5.Items[left2].SubItems[2].Text == listView2.Items[turnout].SubItems[1].Text)      //Check Left2 Turnout state against the actual turnout state
                    {
                        track_piece = listView4.Items[current_track].SubItems[1].Text;
                    }
                    else 
                    {
                        track_piece = listView4.Items[current_track].SubItems[0].Text;
                    }
                }
            }
            else
            {
                //Train should be going right, pick the next track and sensor
                next_sensor = listView4.Items[Convert.ToInt32(listView7.Items[train_location].SubItems[2].Text)].SubItems[9].Text;

                if (listView4.Items[current_track].SubItems[3].Text == "none")       // Get what track they are currently on
                {
                    //Only one track piece can be selected for this direction
                    track_piece = listView4.Items[current_track].SubItems[2].Text;
                }
                else
                {
                    //Multiple pieces could be available - look at turnout state to then see which track it should be going to
                    int right2 = Convert.ToInt32(listView4.Items[current_track].SubItems[3].Text);
                    int turnout = Convert.ToInt32(listView5.Items[right2].SubItems[1].Text);

                    if(listView5.Items[right2].SubItems[2].Text == listView2.Items[turnout].SubItems[1].Text)      //Check Left2 Turnout state against the actual turnout state
                    {
                        track_piece = listView4.Items[current_track].SubItems[3].Text;
                    }
                    else 
                    {
                        track_piece = listView4.Items[current_track].SubItems[2].Text;
                    }
                }
            }
        }

        private void RequestNextTrack(string train_id)
        // ################################################################################################################
        //   Figures out what the next track is that the train is moving and attempts to gain ownership
        // ################################################################################################################
        {
            //Enter monitors
            Monitor.Enter(listView5);              // Resource
            Monitor.Enter(listView2);              // Turnout
            Monitor.Enter(listView7);              // Train

            string track_piece_string = "0";
            int track_piece;
            string next_sensor_string = "0";
            int next_sensor;
            int train_location = GetTrainLocation(train_id);
            int current_track = Convert.ToInt32(listView7.Items[train_location].SubItems[2].Text);
            bool available = true;

            //See if the user already has a requested track
            DeallocateRequestTrack(train_location);
            GetRequestedTrack(train_location, current_track, ref track_piece_string, ref next_sensor_string);

            if (track_piece_string == "dead")
            {
                //Going to be a dead-end
                listView7.Items[train_location].SubItems[5].Text = "0";                   //Set the speed to 0

                TrainForward(Convert.ToInt32(train_id), Convert.ToInt32(listView7.Items[train_location].SubItems[5].Text));    //Stop the train

                listView7.Items[train_location].SubItems[4].Text = "true";                //Not available - block train

                Monitor.Exit(listView7);              // Train
                Monitor.Exit(listView2);              // Turnout
                Monitor.Exit(listView5);              // Resource

                return;
            }

            track_piece = Convert.ToInt32(track_piece_string);
            next_sensor = Convert.ToInt32(next_sensor_string);

            //See if turnout is associated with this piece and if it's turned the right way
            if (listView4.Items[track_piece].SubItems[4].Text == "none")
            {
                //No turnout associated with this track - see if it's available
                if (listView5.Items[track_piece].SubItems[0].Text == "none")
                {
                    //Track is available because of no owner - keep Available true
                }
                else
                {
                    available = false;                       //Owned or dead-end
                }
            }
            else
            {
                int turnout_id = Convert.ToInt32(listView4.Items[track_piece].SubItems[4].Text);

                if ((listView2.Items[turnout_id].SubItems[1].Text == listView5.Items[track_piece].SubItems[2].Text)     //Check turnout state of turnout and resource structure
                    && (listView5.Items[track_piece].SubItems[0].Text == "none"))
                {
                    //Track is available - set turnout and keep Available at true
                    listView2.Items[turnout_id].SubItems[2].Text = train_id.ToString();         //Set it in turnout
                }
                else
                {
                    available = false;
                }
            }

            if (available == false)
            {
                listView7.Items[train_location].SubItems[5].Text = "0";                   //Set the speed to 0

                TrainForward(Convert.ToInt32(train_id), Convert.ToInt32(listView7.Items[train_location].SubItems[5].Text));    //Stop the train

                listView7.Items[train_location].SubItems[4].Text = "true";                //Not available - block train
            }
            else
            {
                listView7.Items[train_location].SubItems[3].Text = track_piece.ToString();  //Set it in train
                listView5.Items[track_piece].SubItems[0].Text = train_id.ToString();        //Set it in resource

                string imagename = listView7.Items[train_location].SubItems[3].Text + "_" + listView7.Items[train_location].SubItems[12].Text + "_light";
                track[Convert.ToInt32(listView7.Items[train_location].SubItems[3].Text)].Image = (Image)Properties.Resources.ResourceManager.GetObject(imagename);
                track[Convert.ToInt32(listView7.Items[train_location].SubItems[3].Text)].Visible = true;

                listView7.Items[train_location].SubItems[4].Text = "false";                  //Set train as not being blocked
                listView7.Items[train_location].SubItems[10].Text = next_sensor.ToString();  //Set next sensor
            }

            //Exit monitors
            Monitor.Exit(listView7);              // Train
            Monitor.Exit(listView2);              // Turnout
            Monitor.Exit(listView5);              // Resource
        }

        private void TrainForward(int train_id, int speed)
        // ################################################################################################################
        //   Increases the speed 'forward' for the train
        // ################################################################################################################
        {
            byte[] command = new byte[6];
            byte[] buffer = new byte[2];

            if (speed != 0)
            {
                speed = speed + 40;                     //Offset of 40 used as speed is VERY slow otherwise
            }

            command[0] = 0xA2;                    //Standard command for motors
            command[1] = 0;
            command[2] = (byte)train_id;          //Train ID
            command[3] = 02;                      //Forward direction
            command[4] = (byte)speed;             //Speed to send command

            try
            {
                port.Write(command, 0, 5);            //Write to the port

                port.Read(buffer, 0, 1);
            }
            catch { }
        }

        private void TrainBackward(int train_id, int speed)
        // ################################################################################################################
        //   Increases the speed 'backwards' for the train
        // ################################################################################################################
        {
            byte[] command = new byte[6];
            byte[] buffer = new byte[2];

            if (speed != 0)
            {
                speed = speed - 40;
            }

            command[0] = 0xA2;                    //Standard command for motors
            command[1] = 0;
            command[2] = (byte)train_id;          //Train ID
            command[3] = 01;                      //Backward direction
            command[4] = (byte)Math.Abs(speed);   //Speed to send command

            try
            {
                port.Write(command, 0, 5);            //Write to the port

                port.Read(buffer, 0, 1);
            }
            catch { }
        }


// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//     BEGINNING OF CODE FOR AIU Processing and changes
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

        private void SensorHitChange(string train_id)
        // ################################################################################################################
        //   Train hit their next expected sensor - change structure information and images
        // ################################################################################################################
        {
            //Enter monitors
            Monitor.Enter(listView5);
            Monitor.Enter(listView2);
            Monitor.Enter(listView7);

            int train_location = GetTrainLocation(train_id);

            DeallocateTrack(train_location);                                                                          //Deallocate current track user is on

            //View requested track of this train and make it it's new owned track
            listView7.Items[train_location].SubItems[2].Text = listView7.Items[train_location].SubItems[3].Text;      //Change 'Track ID' to new
            listView7.Items[train_location].SubItems[3].Text = "none";                                                //Change 'Request' to none

            //Change image on tracked_id associated with this train with the proper color
            string imagename = listView7.Items[train_location].SubItems[2].Text + "_" + listView7.Items[train_location].SubItems[12].Text;
            track[Convert.ToInt32(listView7.Items[train_location].SubItems[2].Text)].Image = (Image)Properties.Resources.ResourceManager.GetObject(imagename);
            track[Convert.ToInt32(listView7.Items[train_location].SubItems[2].Text)].Visible = true;

            //Exit monitors
            Monitor.Exit(listView7);
            Monitor.Exit(listView2);
            Monitor.Exit(listView5);

            RequestNextTrack(train_id);
        }

        private void CheckAIU()
        // ################################################################################################################
        //   Gets the current AIU status from the 'Blackbox' to see if any sensors were hit
        //      > Ideas and bit isolation taken from Chris Kratz
        // ################################################################################################################
        {
            byte[] command = new byte[4];
            int[] value = new int[15];
            byte[] buffer = new byte[2];
            byte[] buffer2 = new byte[2];

            command[0] = 0x9B;                            //Standard command for AIU messages

            try
            {
                for (int i = 0; i < aiu_count; i++)
                {
                    command[1] = (byte)aiu[i];

                    port.Write(command, 0, 2);

                    try
                    {
                        port.Read(buffer, 0, 1);                  // Reads the hi byte
                        port.Read(buffer2, 0, 1);                 // Reads the low byte
                    }
                    catch { }

                    for (int j = 0; j < 6; j++)
                    {
                        byte test = buffer[0];
                        test = (byte)(test >> j);             // Isolate low bytes
                        value[j + 8] = (test & 0x01);
                    }

                    for (int j = 0; j < 8; j++)
                    {
                        byte test = buffer2[0];
                        test = (byte)(test >> j);             // Isolate high bytes
                        value[j] = (test & 0x01);
                    }

                    ChangeAIUImages(value, aiu[i], i);        // Changes images if items changed
                }
            }
            catch
            {
                timer1.Enabled = false;
                timer2.Enabled = false;
                MessageBox.Show("The COM port has been disconnected.  Shutting down the server.");
                Form1_FormClosing(null, null);
            }
        }

        private void ChangeandRequestAIU(int sensor_id)
        // ################################################################################################################
        //   Make sensor visible and look through Next Sensor Id's for Trains to see if it matches
        // ################################################################################################################
        {
            sensors[sensor_id].Visible = true;

            for (int i = 0; i < listView7.Items.Count; i++)
            {
                if ((listView7.Items[i].SubItems[10].Text == sensor_id.ToString()))                      // See if sensor is under next sensor in Train structure
                {
                    if ((listView7.Items[i].SubItems[4].Text == "false") && (listView7.Items[i].SubItems[1].Text != "none"))  // Now check if it is blocked and has an owner
                    {
                        SensorHitChange(listView7.Items[i].SubItems[0].Text);                            // Move if available
                    }
                }
            }
        }

        private void ChangeAIUImages(int[] value, int aiu_nbr, int counter)
        // ################################################################################################################
        //   Logic to check AIU current status with default to change images to show 'hits'
        // ################################################################################################################
        {
            switch (aiu_nbr)
            {
                case 50:
                    {
                        if (listView9.Items[counter].SubItems[3].Text != value[3].ToString()) { ChangeandRequestAIU(16); }
                        else { sensors[16].Visible = false; }

                        if (listView9.Items[counter].SubItems[4].Text != value[4].ToString()) { ChangeandRequestAIU(23); }
                        else { sensors[23].Visible = false; }

                        if (listView9.Items[counter].SubItems[5].Text != value[5].ToString()) { ChangeandRequestAIU(21); }
                        else { sensors[21].Visible = false; }

                        if (listView9.Items[counter].SubItems[6].Text != value[6].ToString()) { ChangeandRequestAIU(2); }
                        else { sensors[2].Visible = false; }

                        if (listView9.Items[counter].SubItems[7].Text != value[7].ToString()) { ChangeandRequestAIU(1); }
                        else { sensors[1].Visible = false; }

                        if (listView9.Items[counter].SubItems[8].Text != value[8].ToString()) { ChangeandRequestAIU(3); }
                        else { sensors[3].Visible = false; }

                        if (listView9.Items[counter].SubItems[9].Text != value[9].ToString()) { ChangeandRequestAIU(22); }
                        else { sensors[22].Visible = false; }

                        if (listView9.Items[counter].SubItems[10].Text != value[10].ToString()) { ChangeandRequestAIU(15); }
                        else { sensors[15].Visible = false; }

                        if (listView9.Items[counter].SubItems[11].Text != value[11].ToString()) { ChangeandRequestAIU(4); }
                        else { sensors[4].Visible = false; }

                        if (listView9.Items[counter].SubItems[12].Text != value[12].ToString()) { ChangeandRequestAIU(17); }
                        else { sensors[17].Visible = false; }

                        if (listView9.Items[counter].SubItems[13].Text != value[13].ToString()) { ChangeandRequestAIU(0); }
                        else { sensors[0].Visible = false; }

                        break;
                    }
                case 58:
                    {
                        if (listView9.Items[counter].SubItems[0].Text != value[0].ToString()) { ChangeandRequestAIU(14); }
                        else { sensors[14].Visible = false; }

                        if (listView9.Items[counter].SubItems[1].Text != value[1].ToString()) { ChangeandRequestAIU(19); }
                        else { sensors[19].Visible = false; }

                        if (listView9.Items[counter].SubItems[2].Text != value[2].ToString()) { ChangeandRequestAIU(26); }
                        else { sensors[26].Visible = false; }

                        if (listView9.Items[counter].SubItems[3].Text != value[3].ToString()) { ChangeandRequestAIU(20); }
                        else { sensors[20].Visible = false; }

                        if (listView9.Items[counter].SubItems[4].Text != value[4].ToString()) { ChangeandRequestAIU(24); }
                        else { sensors[24].Visible = false; }

                        if (listView9.Items[counter].SubItems[5].Text != value[5].ToString()) { ChangeandRequestAIU(10); }
                        else { sensors[10].Visible = false; }

                        if (listView9.Items[counter].SubItems[6].Text != value[6].ToString()) { ChangeandRequestAIU(13); }
                        else { sensors[13].Visible = false; }

                        if (listView9.Items[counter].SubItems[7].Text != value[7].ToString()) { ChangeandRequestAIU(9); }
                        else { sensors[9].Visible = false; }

                        if (listView9.Items[counter].SubItems[8].Text != value[8].ToString()) { ChangeandRequestAIU(7); }
                        else { sensors[7].Visible = false; }

                        if (listView9.Items[counter].SubItems[9].Text != value[9].ToString()) { ChangeandRequestAIU(11); }
                        else { sensors[11].Visible = false; }

                        if (listView9.Items[counter].SubItems[10].Text != value[10].ToString()) { ChangeandRequestAIU(6); }
                        else { sensors[6].Visible = false; }

                        if (listView9.Items[counter].SubItems[11].Text != value[11].ToString()) { ChangeandRequestAIU(5); }
                        else { sensors[5].Visible = false; }

                        if (listView9.Items[counter].SubItems[12].Text != value[12].ToString()) { ChangeandRequestAIU(25); }
                        else { sensors[25].Visible = false; }

                        if (listView9.Items[counter].SubItems[13].Text != value[13].ToString()) { ChangeandRequestAIU(8); }
                        else { sensors[8].Visible = false; }

                        break;
                    }
                case 62:
                    {
                        if (listView9.Items[counter].SubItems[3].Text != value[3].ToString()) { ChangeandRequestAIU(28); }
                        else { sensors[28].Visible = false; }

                        if (listView9.Items[counter].SubItems[4].Text != value[4].ToString()) { ChangeandRequestAIU(18); }
                        else { sensors[18].Visible = false; }

                        if (listView9.Items[counter].SubItems[5].Text != value[5].ToString()) { ChangeandRequestAIU(27); }
                        else { sensors[27].Visible = false; }

                        if (listView9.Items[counter].SubItems[6].Text != value[6].ToString()) { ChangeandRequestAIU(12); }
                        else { sensors[12].Visible = false; }

                        break;
                    }
            }
        }


// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//     BEGINNING OF CODE FOR CHANGING IMAGES
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

        private void pictureBox72_MouseHover(object sender, EventArgs e)
        // ################################################################################################################
        //   Mouse is over the Add Train button, change to highlighted
        // ################################################################################################################
        {
            pictureBox72.Image = Properties.Resources.add_train_over;
        }
        private void pictureBox76_MouseHover(object sender, EventArgs e)
        // ################################################################################################################
        //   Mouse is over the Remove button, change to highlighted
        // ################################################################################################################
        {
            pictureBox76.Image = Properties.Resources.remove_over;
        }

        private void pictureBox72_MouseLeave(object sender, EventArgs e)
        // ################################################################################################################
        //   Mouse has left the Add Train button, change to unhighlighted
        // ################################################################################################################
        {
            pictureBox72.Image = Properties.Resources.add_train;
        }
        private void pictureBox76_MouseLeave(object sender, EventArgs e)
        // ################################################################################################################
        //   Mouse has left the Remove button, change to unhighlighted
        // ################################################################################################################
        {
            pictureBox76.Image = Properties.Resources.remove;
        }


// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//     BEGINNING OF CODE FOR HANDLING CLOSING AND DISCONNECTS
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

        private void CloseSocket(Socket TempSocket)
        // ################################################################################################################
        //   User has disconnected - open up the train for another user to use
        // ################################################################################################################
        {
            for (int i = 0; i < listView1.Items.Count; i++)
            {
                if (TempSocket.RemoteEndPoint.ToString() == listView1.Items[i].SubItems[2].Text)
                {
                    listView1.Items[i].SubItems[1].Text = "";          // Open up this for another user
                    listView1.Items[i].SubItems[2].Text = "Available";

                    for (int j = 0; j < listView7.Items.Count; j++)
                    {
                        if (listView7.Items[j].SubItems[0].Text == listView1.Items[i].SubItems[0].Text)
                        {
                            listView7.Items[j].SubItems[1].Text = "none";                                    // Open up ownership in Train area also
                            listView7.Items[j].SubItems[5].Text = "0";                                       // Set the speed to 0
                            TrainForward(Convert.ToInt32(listView7.Items[j].SubItems[0].Text), 0);           // Stop the train
                        }
                    }

                    listBox1.Invoke(new WritetoLogsCallback(WritetoLogs), DateTime.Now.ToString("MM/dd/yyyy hh:mm:ss") + " - Client " + TempSocket.RemoteEndPoint.ToString() + " disconnected from train " + listView1.Items[i].SubItems[0].Text);

                    TempSocket.Close();
                    TempSocket = null;

                    connection_count--;                                // Decrease connection count

                    label2.Invoke(new UpdateLabelsCallback(UpdateLabels), label2, "Connections: " + connection_count);

                    return;
                }
            }
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        // ################################################################################################################
        //   Form is closing - close any potentially open sockets
        // ################################################################################################################
        {
            try
            {
                if (ServerSocket != null)
                {
                    ServerSocket.Close();
                }
                for (int i = 0; i < total_connections; i++)     //Go through connection count, close client sockets
                {
                    TrainForward(Convert.ToInt32(listView1.Items[i].SubItems[0].Text), 0);        // Stop all of the trains
                    Clients[Convert.ToInt32(listView1.Items[i].SubItems[1].Text)].Close();        //Close these out also
                }
            }
            catch
            {
                //MessageBox.Show("Issue closing Server Socket and clients");
            }

            try
            {
                port.Close();                                   // Close serial port
            }
            catch
            {
                //MessageBox.Show("Problem Closing Thread test.exe " + exc.ToString());
            }

            Application.Exit();
        }


// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//     BEGINNING OF CODE FOR LOGS AND ACTIONS
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

        private void button1_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   'Save Logs' button has been chlicked - display savefiledialog and save
        // ################################################################################################################
        {
            string current_date = DateTime.Now.ToString("MMddyyyy");

            saveFileDialog1.FileName = "C:\\trainlog_" + current_date + ".txt";      // Default file name
            if (saveFileDialog1.ShowDialog() == DialogResult.Cancel)
            {
                return;
            }
            string filename = saveFileDialog1.FileName;

            //----- Create the output file
            FileStream file = new FileStream(filename, FileMode.OpenOrCreate, FileAccess.Write);
            StreamWriter sw = new StreamWriter(file);

            // Write log information to file
            for (int i = 0; i < listBox1.Items.Count; i++)
            {
                sw.WriteLine(listBox1.Items[i].ToString());
            }

            sw.Close();
            file.Close();
        }

        private void button7_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   Do a simple 'Print Preview' to print out at least the first page of the log information
        //      > Modified from an example of Dr. Blahnik's
        // ################################################################################################################
        {
            pd = new PrintDocument();
            pd.PrinterSettings = pageSetupDialog1.PrinterSettings;
            pd.DefaultPageSettings.Landscape = false;                                 // Setting up print preview
            pd.PrintPage += new PrintPageEventHandler(pd_PrintPage);
            PrintPreviewDialog dlg = new PrintPreviewDialog();
            dlg.Document = pd;
            dlg.ShowDialog();                                                         // Actually Show Preview
        }

        private void pictureBox64_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   Expand to show logs 'button' clicked - expand or limit
        // ################################################################################################################
        {
            if (expanded == false)
            {
                Form1.ActiveForm.Width = 1024;
                Form1.ActiveForm.Height = 865;                                 // Change size and image
                pictureBox64.Image = Properties.Resources.expand_logs_up;
                expanded = true;
            }
            else
            {
                Form1.ActiveForm.Width = 1024;
                Form1.ActiveForm.Height = 720;                                 // Change size and image
                pictureBox64.Image = Properties.Resources.expand_logs;
                expanded = false;
            }
        }

        public void WritetoLogs(string loginfo)
        // ################################################################################################################
        //   Standard delegate to use to write logs to listBox1 on main thread
        // ################################################################################################################
        {
            listBox1.Items.Add(loginfo);
        }

        private void pd_PrintPage(object sender, PrintPageEventArgs ev)
        // #####################################################################################
        //   Populate print preview for printing on the outputs
        // #####################################################################################
        {
            int i = 0;
            int height = 20;
            System.Drawing.Font ff = new Font("Times New Roman", 10);

            for (; i < listBox1.Items.Count; i++)
            {
                ev.Graphics.DrawString(listBox1.Items[i].ToString(), ff, Brushes.Black, 75, height);

                height = height + 12;

                if (i == 100)
                    break;
            }

            if (i == 0)
            {
                ev.HasMorePages = true;
            }
            else
            {
                ev.HasMorePages = false;
            }
        }

        private void button8_Click(object sender, EventArgs e)
        // ################################################################################################################
        //   Standard print setup dialog that allows a user to set default printer, etc.
        // ################################################################################################################
        {
            pageSetupDialog1.PageSettings = new System.Drawing.Printing.PageSettings();

            pageSetupDialog1.PrinterSettings = new System.Drawing.Printing.PrinterSettings();     // Initialize dialog's PrinterSettings property 
            pageSetupDialog1.ShowNetwork = false;                                                 // No network printers

            pageSetupDialog1.ShowDialog();
        }
    }
}